/*
 * Copyright (C), 2000-2002 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#ifdef SOLARIS
#include <procfs.h>
#endif

#ifdef FREEBSD
#include <kvm.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/resource.h>
#include <vm/vm_object.h>
#include <vm/vm_map.h>
#include <sys/vmmeter.h>
#include <sys/sysctl.h>
#include <vm/vm.h>
#endif

#ifdef LINUX
#include <asm/param.h>
#include <asm/page.h>
#endif


#include "process.h"

/**
 *  System dependent /proc & kvm methods.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id$
 *
 *  @file
 */

/* ------------------------------------------------------------------ Public */

#ifdef SOLARIS

int init_process_info_sysdep(void) {

  if( getuid()!=0 ) {

    error("%s: You must run monit as root to utilize the "
	  "process status functionality.\n", prog);
    return FALSE;

  }

  return TRUE;

}

double timestruc_to_tseconds(timestruc_t t) {
  return  t.tv_sec * 10 + t.tv_nsec / 100000000.0;
}

int get_process_info_sysdep(Proc_Info_T p) {

  char buf[4096];
  psinfo_t  * psinfo=  (psinfo_t *)&buf;
  pstatus_t * pstatus= (pstatus_t *)&buf;

  if (!read_proc_file(buf,4096, "psinfo", p->pid)) {

    return FALSE;

  }

  /* If we don't have any light-weight processes (LWP) then we
     are definitely a zombie */

  if ( psinfo->pr_nlwp == 0 ) {

    p->status_flag = PROCESS_ZOMBIE;

  }


  if ( p->status_flag != PROCESS_ZOMBIE ) {
    /* We can't access /proc/$pid/status of a zombie */
    /* and does it anyway matter? */

    p->mem_percent = psinfo->pr_pctmem * 1000 / 0x8000;
    p->mem_kbyte = psinfo->pr_rssize;

    if (!read_proc_file(buf,4096, "status", p->pid)) {

      return FALSE;

    }

    p->cputime_prev= p->cputime;
    p->cputime= ( timestruc_to_tseconds(pstatus->pr_utime) +
		  timestruc_to_tseconds(pstatus->pr_stime) );

    if( include_children ) {

      p->cputime+= ( timestruc_to_tseconds(pstatus->pr_cutime) +
		     timestruc_to_tseconds(pstatus->pr_cstime) );

    }
    /* first run ? */

    if ( p->time_prev == 0.0 ) {

      p->cputime_prev= p->cputime;

    }

  } else {

    p->cputime_prev= p->cputime = 0;
    p->mem_kbyte= 0;
    p->mem_percent= 0.0;

  }

  return TRUE;
}

#endif

#ifdef LINUX

#define PAGE_TO_KBYTE_SHIFT PAGE_SHIFT-10

static long mem_kbyte_max;

int init_process_info_sysdep(void) {

  struct stat buf;

  /* I hope that hack to get the memsize is okay. (-: */

  if ( stat("/proc/kcore", &buf) != 0 ) {

    return FALSE;

  }

  mem_kbyte_max = buf.st_size>>10;

  return TRUE;

}

int get_process_info_sysdep(Proc_Info_T p) {

  char buf[4096];
  char* tmp;
  char stat_item_state;
  unsigned long stat_item_utime;
  unsigned long stat_item_stime;
  long stat_item_cutime;
  long stat_item_cstime;
  long stat_item_rss;

  if (!read_proc_file(buf,4096, "stat", p->pid)) {

    return FALSE;

  }

  /* Move along the buffer to get past the process name */

  tmp = strrchr(buf, ')') + 2;

  /* This implementation is done by using fs/procfs/array.c as a basis
     it is also worth looking into the source of the procps utils */

  sscanf(tmp,"%c %*d %*d %*d %*d %*d %*lu %*lu"
	 "%*lu %*lu %*lu %lu %lu %ld %ld %*ld %*ld %*ld "
	 "%*ld %*lu %*lu %ld %*lu %*lu %*lu %*lu %*lu "
	 "%*lu %*lu %*lu %*lu %*lu %*lu %*lu %*lu %*d %*d\n",
	 &stat_item_state, &stat_item_utime, &stat_item_stime,
	 &stat_item_cutime, &stat_item_cstime, &stat_item_rss);

  /* abs to please the compiler... we dont want to shift negatively.
     why doesn't C understand this??? */

  if ( PAGE_TO_KBYTE_SHIFT < 0 ) {

    p->mem_kbyte= stat_item_rss >> abs(PAGE_TO_KBYTE_SHIFT);

  } else {

    p->mem_kbyte= stat_item_rss << abs(PAGE_TO_KBYTE_SHIFT);

  }

  p->mem_percent = (int) ((double) p->mem_kbyte * 1000.0 / mem_kbyte_max);

  /* jiffies -> seconds = 1 / HZ
     HZ is defined in "asm/param.h"  and it is usually 1/100s but on
     alpha system it is 1/1024s */

  p->cputime_prev = p->cputime;
  p->cputime =  ( stat_item_utime + stat_item_stime ) * 10 / HZ;

  if ( include_children ) {

    p->cputime += ( stat_item_cutime + stat_item_cstime ) * 10 / HZ;

  }

  /* first run ? */

  if ( p->time_prev == 0.0 ) {

      p->cputime_prev = p->cputime;

  }

  /* State is Zombie -> then we are a Zombie ... clear or? (-: */

  if ( stat_item_state == 'Z' ) {

    p->status_flag |= PROCESS_ZOMBIE;

  }

  return TRUE;

}

#endif

#ifdef FREEBSD

#define pagetok(size) ((size) << pageshift)

static int pageshift;
static long mem_kbyte_max;

#ifndef LOG1024
#define LOG1024         10
#endif

static kvm_t * kvm_handle;

int init_process_info_sysdep(void) {

  register int pagesize;
  struct vmmeter vmm;

  kvm_handle = kvm_open(NULL, NULL, NULL, O_RDONLY, argv[0]);

  if ( kvm_handle == NULL ) {

    return FALSE;

  }

  /* ----------------------------- INIT MEM -----------------------------*/
  /* Got it from libgtop/sysdep/freebsd/mem.c */

  /* get the page size with "getpagesize" and calculate pageshift
   * from it */
  pagesize = getpagesize ();
  pageshift = 0;
  while (pagesize > 1) {

    pageshift++;
    pagesize >>= 1;

  }
  
  /* we only need the amount of log(2)1024 for our conversion */
  pageshift -= LOG1024;

  /* Get the data from kvm_* */
  if (kvm_read (kvm_handle, nlst[1].n_value,
		&vmm, sizeof (vmm)) != sizeof (vmm)) {

    return FALSE;

  }

  mem_kbyte_max= vmm.v_kernel_pages +
    vmm.v_free_count + vmm.v_wire_count +
    vmm.v_active_count + vmm.v_inactive_count;

  return TRUE;

}

int get_process_info_sysdep(Proc_Info_T p) {

  struct kinfo_proc *pinfo;
  struct user *u_addr = (struct user *)USRSTACK;
  struct pstats pstats;
  struct plimit plimit;
  struct vmspace *vms;
  register struct rusage *rup;
  long stat_utime;
  long stat_stime;
  long stat_cutime;
  long stat_cstime;

  u_int64_t stat_rss;
  u_int64_t rss_lim;

  int count;

  /* Got it from libgtop */

  pinfo = kvm_getprocs(kvm_handle, KERN_PROC_PID, p->pid, &count);

  if ((pinfo == NULL) || (count < 1)) {

    return FALSE;

  }

  /* ----------------------------- CPU TIMING ----------------------------*/
  /* Got it from libgtop/sysdep/freebsd/proctime.c */

  if ((pinfo [0].kp_proc.p_flag & P_INMEM) &&
      kvm_uread (kvm_handle, &(pinfo [0]).kp_proc,
		 (unsigned long) &u_addr->u_stats,
		 (char *) &pstats, sizeof (pstats)) == sizeof (pstats)) {

    rup = &pstats.p_ru;
    calcru(&(pinfo [0]).kp_proc,
	   &rup->ru_utime, &rup->ru_stime, NULL);

    stat_utime = tv2sec (pstats.p_ru.ru_utime);
    stat_stime = tv2sec (pstats.p_ru.ru_stime);

    stat_cutime = tv2sec (pstats.p_cru.ru_utime);
    stat_cstime = tv2sec (pstats.p_cru.ru_stime);

  } else {

    return FALSE;

  }

  p->cputime_prev= p->cputime;
  p->cputime= ( stat_utime + stat_stime );

  if( include_children ) {

    p->cputime+= ( stat_cutime + stat_cstime );

  }

  /* first run ? */

  if ( p->time_prev == 0.0 ) {

    p->cputime_prev= p->cputime;

  }

  /* ----------------------------- MEMORY --------------------------------*/
  /* Got it from libgtop/sysdep/freebsd/procmem.c */

  if (kvm_read (kvm_handle,
		(unsigned long) pinfo [0].kp_proc.p_limit,
		(char *) &plimit, sizeof (plimit)) != sizeof (plimit)) {

    return FALSE;

  }

  rss_rlim = (u_int64_t)
    (plimit.pl_rlimit [RLIMIT_RSS].rlim_cur);

  vms = &pinfo [0].kp_eproc.e_vm;

  p->mem_kbyte= (u_int64_t) pagetok (vms->vm_rssize);
             /* <<LOG1024 removed, we wanna have kb */

  /* ----------------------------- STATE ---------------------------------*/
  /* Got it from libgtop/sysdep/freebsd/procstate.c */

  if ( pinfo [0].kp_proc.p_stat == SZOMB ) {

    p->status_flag |= PROCESS_ZOMBIE;

  }

  p->mem_percent = (int) ((double) p->mem_kbyte * 1000.0 / mem_kbyte_max);

  return TRUE;

}

#endif

